__author__ = 'aymgal'

from coolest.template.classes.regularization import Regularization
from coolest.template.classes.parameter import HyperParameter
from coolest.template.classes.parameter import DefinitionRange


__all__ = [
    'PixelEntropy',
    'PixelRidge',
    'PixelLasso',
    'PixelGradient',
    'PixelCurvature',
    'PixelExponentialKernel',
    'PixelGaussianKernel',
    'PixelStarlet',
    'PixelBLWavelet',
    'PixelPositivity',
    'DefaultForPixelated',
]
SUPPORTED_CHOICES = __all__


class PixelEntropy(Regularization):

    def __init__(self, **kwargs):
        description = "Entropy regularization"
        parameters = {
            'lambda': HyperParameter("Lagrange parameter for entropy regularization",
                                     DefinitionRange(min_value=0.0),
                                     latex_str=r"$\lambda_{\rm e}$"),
            }
        super().__init__(description, parameters, **kwargs)


class PixelRidge(Regularization):

    def __init__(self, **kwargs):
        description = "Ridge regularization (L2 norm)"
        parameters = {
            'lambda': HyperParameter("Lagrange parameter for Ridge norm regularization",
                                     DefinitionRange(min_value=0.0),
                                     latex_str=r"$\lambda_{\rm ridge}$"),
        }
        super().__init__(description, parameters, **kwargs)


class PixelLasso(Regularization):

    def __init__(self, **kwargs):
        description = "Lasso regularization (L1 norm)"
        parameters = {
            'lambda': HyperParameter("Lagrange parameter for Lasso norm regularization",
                                     DefinitionRange(min_value=0.0),
                                     latex_str=r"$\lambda_{\rm lasso}$"),
        }
        super().__init__(description, parameters, **kwargs)


class PixelGradient(Regularization):

    def __init__(self, **kwargs):
        description = "Gradient regularization (first spatial derivative)"
        parameters = {
            'lambda': HyperParameter("Lagrange parameter for gradient regularization",
                           DefinitionRange(min_value=0.0),
                           latex_str=r"$\lambda_{\rm grad}$"),
        }
        super().__init__(description, parameters, **kwargs)


class PixelCurvature(Regularization):

    def __init__(self, **kwargs):
        description = "Curvature regularization (second spatial derivative)"
        parameters = {
            'lambda': HyperParameter("Lagrange parameter for curvature regularization",
                           DefinitionRange(min_value=0.0),
                           latex_str=r"$\lambda_{\rm curv}$"),
        }
        super().__init__(description, parameters, **kwargs)


class PixelExponentialKernel(Regularization):

    def __init__(self, **kwargs):
        description = "Exponential kernel regularization"
        parameters = {
            'lambda': HyperParameter("Lagrange parameter for exponential kernel regularization",
                           DefinitionRange(min_value=0.0),
                           latex_str=r"$\lambda_{\rm exp}$"),
        }
        super().__init__(description, parameters, **kwargs)


class PixelGaussianKernel(Regularization):

    def __init__(self, **kwargs):
        description = "Gaussian kernel regularization"
        parameters = {
            'lambda': HyperParameter("Lagrange parameter for gaussian kernel regularization",
                           DefinitionRange(min_value=0.0),
                           latex_str=r"$\lambda_{\rm gauss}$"),
        }
        super().__init__(description, parameters, **kwargs)


class PixelStarlet(Regularization):

    def __init__(self, **kwargs):
        description = "Starlets regularization"
        parameters = {
            'lambda_hf': HyperParameter("Lagrange parameter for first starlet scale (high frequencies)",
                           DefinitionRange(min_value=0.0),
                           latex_str=r"$\lambda_{\rm s, hf}$"),
            'lambda': HyperParameter("Lagrange parameter from second to last starlet scales (excluding coarse scale)",
                           DefinitionRange(min_value=0.0),
                           latex_str=r"$\lambda_{\rm s}$"),
        }
        super().__init__(description, parameters, **kwargs)


class PixelBLWavelet(Regularization):

    def __init__(self, **kwargs):
        description = "Battle-Lemarie wavelet regularization"
        parameters = {
            'lambda_hf': HyperParameter("Lagrange parameter for first Battle-Lemarie scale (high frequencies)",
                           DefinitionRange(min_value=0.0),
                           latex_str=r"$\lambda_{\rm BL, hf}$"),
        }
        super().__init__(description, parameters, **kwargs)

class PixelPositivity(Regularization):

    def __init__(self, **kwargs):
        description = "Non-negativity constraint on pixel values"
        parameters = {
            'lambda': HyperParameter("Lagrange parameter",
                           DefinitionRange(min_value=0.0),
                           latex_str=r"$\lambda_{\rm pos}$"),
        }
        super().__init__(description, parameters, **kwargs)


class DefaultForPixelated(Regularization):

    def __init__(self, **kwargs):
        description = "Default regularization strategy, set by the specific modelling code"
        parameters = {
            'lambda': HyperParameter("Lagrange parameter",
                           DefinitionRange(min_value=0.0),
                           latex_str=r"$\lambda$"),
        }
        super().__init__(description, parameters, **kwargs)



__author__ = 'aymgal', 'mattgomer', 'gvernard'

import numpy as np
from astropy.coordinates import SkyCoord
import logging

from coolest.api.composable_models import *
from coolest.api import util

# logging settings
logging.getLogger().setLevel(logging.INFO)

class Analysis(object):
    """Handles computation of model-independent quantities 
    and other analysis computations.
    
    NOTE: Except for methods that do have a `coordinates` keyword argument,
    the grid used to performed the computations will always be the one corresponding
    to the instrument / observation field-of-view, with resolution controlled by the
    `supersampling` keyword argument below.

    Parameters
    ----------
    coolest_object : COOLEST
        COOLEST instance
    coolest_directory : str
        Directory which contains the COOLEST template and other data files
    supersampling : int, optional
        Supersampling factor (relative to the instrument pixel size)
        that defines the grid on which computations are performed, by default 1.
    """

    def __init__(self, coolest_object, coolest_directory, supersampling=1):
        self.coolest = coolest_object
        self.coolest_dir = coolest_directory
        base_coordinates = util.get_coordinates(self.coolest)
        if supersampling > 1:
            self.coordinates = base_coordinates.create_new_coordinates(pixel_scale_factor=1./supersampling)
        else:
            self.coordinates = base_coordinates

    def effective_einstein_radius(self, center=None, initial_guess=1, initial_delta_pix=10, 
                                  n_iter=5, return_accuracy=False, max_loopcount=100, 
                                  **kwargs_selection):
        """Calculates Einstein radius for a kappa grid starting from an initial guess with large step size and zeroing in from there.
        Uses the grid from the create_kappa_image which is built from the coolest file.

        Parameters
        ----------
        center : (float, float), optional
            (x, y)-coordinates of the center from which to calculate Einstein radius; if None, use the value from create_kappa_image, by default None
        initial_guess : int, optional
            initial guess for Einstein radius, by default 1
        initial_delta_pix : int, optional
            initial step size before shrinking in future iterations, by default 10
        n_iter : int, optional
            number of iterations, by default 5
        return_accuracy : bool, optional
            if True, return estimate of accuracy as well as R_Ein value, by default False

        Returns
        -------
        float, or (float, float) if return_accuracy is True
            Effective Einstein radius

        Raises
        ------
        RuntimeError
            If the algorithm is running for more than 100 loops.
        """
        if kwargs_selection is None:
            kwargs_selection = {}
        mass_model = ComposableMassModel(self.coolest, self.coolest_dir, **kwargs_selection)

        # get an image of the convergence
        x, y = self.coordinates.pixel_coordinates
        kappa_image = mass_model.evaluate_convergence(x, y)

        # import matplotlib.pyplot as plt
        # plt.imshow(np.log10(kappa_image))
        # plt.show()

        # select a center
        if center is None:
            center_x, center_y = mass_model.estimate_center()
        else:
            center_x, center_y = center

        def _loop_until_overshoot(r_Ein, delta, direction, runningtotal, total_pix):
            """
            this subfunction iteratively adjusts the mask radius by delta either inward or outward until the sign flips on mean_kappa-area
            """
            loopcount=0
            if r_Ein==np.nan: #return nan if given nan (needed when I put this in a loop below)
                return np.nan, np.nan, 0, runningtotal, total_pix
            while (runningtotal-total_pix)*direction>0:
                if loopcount > max_loopcount:
                    raise RuntimeError('Stuck in very long (possibly infinite) loop')
                if direction > 0:
                    mask=only_between_r1r2(r_Ein, r_Ein+delta,r_grid)
                else:
                    mask=only_between_r1r2(r_Ein-delta, r_Ein, r_grid)
                kappa_annulus=np.sum(kappa_image[mask])
                added_pix=np.sum(mask)
                runningtotal=runningtotal+kappa_annulus*direction
                total_pix=total_pix+added_pix*direction
                r_Ein=r_Ein+delta*direction

                if r_Ein < min_guess:
                    #Either we overshot into the center region between pixels (rare), or kappa is subcritical.
                    #check starting from zero
                    mask=only_between_r1r2(0,min_guess,r_grid)
                    runningtotal=np.sum(kappa_image[mask])
                    total_pix=np.sum(mask)
                    if runningtotal-total_pix < 0:
                        logging.warning('kappa is sub-critical, Einstein radius undefined.')
                        return np.nan, np.nan, 0, runningtotal, total_pix
                loopcount+=1
            return r_Ein, delta, direction, runningtotal, total_pix

        def only_between_r1r2(r1, r2, r_grid):
            if r1 > r2:
                raise ValueError('r2 must be greater than r1')
            return (r_grid >= r1) & (r_grid < r2)

        #some initialization: if there is an even number of pixels we don't want to be stuck in the center between 4 pixels.
        grid_res=np.abs(x[0,0]-x[0,1])
        initial_delta=grid_res*initial_delta_pix #default inital step size is 10 pixels
        min_guess=grid_res*np.sqrt(2)+1e-6
        initial_guess=max(initial_guess,min_guess) #initial guess must include at least one pixel

        r_grid=np.sqrt((x-center_x)**2+(y-center_y)**2)
        mask=only_between_r1r2(0,initial_guess,r_grid)
        runningtotal=np.sum(kappa_image[mask])
        total_pix=np.sum(mask)
        if runningtotal > total_pix: #move outward
            direction=1
        elif runningtotal < total_pix: #move inward
            direction=-1
        else:
            return initial_guess
        r_Ein=initial_guess
        delta=initial_delta

        for n in range(n_iter):
            #overshoots, turn around and backtrack at higher precision
            r_Ein, delta, direction, runningtotal, total_pix = _loop_until_overshoot(r_Ein, delta, direction, runningtotal, total_pix)
            direction=direction*-1
            delta=delta/2
        accuracy=grid_res/2 #after testing, accuracy is about grid_res/2
        if np.isnan(r_Ein):
            accuracy=np.nan

        if return_accuracy:
            return r_Ein, accuracy
        else:
            return r_Ein

    def kappa_1d_profile(self, center=None, r_vec=np.linspace(0, 10, 100),**kwargs_selection):
        """Calculates 1D profile using the kappa grid.

        Parameters
        ----------
        center : (float, float), optional
            (x, y)-coordinates of the center from which to calculate; if None, use the value from , by default None
        r_vec : _type_, optional
            range of radii over which to calculate the 1D profile, by default np.linspace(0, 10, 100)

        Returns
        -------
        (array, array)
            kappa values and associated radius values
        """
        if kwargs_selection is None:
            kwargs_selection = {}
        mass_model = ComposableMassModel(self.coolest, self.coolest_dir, **kwargs_selection)
        x, y = self.coordinates.pixel_coordinates
        kappa_image = mass_model.evaluate_convergence(x, y)
        if center is None:
            center_x, center_y = mass_model.estimate_center()
        else:
            center_x, center_y = center
        r_grid=np.sqrt((x-center_x)**2+(y-center_y)**2)
        kappa_profile=np.zeros_like(r_vec)
        for r_i,r in enumerate(r_vec):
            if r_i==0:
                r_min=0
            else:
                r_min=r_vec[r_i-1]
            in_radial_bin=(r_grid > r_min) & (r_grid <= r)
            kappa_profile[r_i]=np.mean(kappa_image[in_radial_bin])
        return kappa_profile, r_vec

    def effective_radial_slope(self, r_eval=None, center=None, r_vec=np.linspace(0, 10, 100),**kwargs_selection):
        """Numerically calculates slope of the kappa profile. Because this is defined on a grid, it is not as accurate or robust as an analytical calculation. 

        Parameters
        ----------
        r_eval : float, optional
            radius at which to return a single value of the slope (e.g. Einstein radius). If None (default), returns slope for all values in r_vec, by default None
        center : (float, float), optional
            (x, y)-coordinates of the center from which to calculate; if None, use the value from create_kappa_image, by default None
        r_vec : array-like, optional
            range of radii over which to calculate the 1D profile, by default np.linspace(0, 10, 100)

        Returns
        -------
        float
            Effective slope
        """
        kappa_profile, r_vec =self.kappa_1d_profile(center=center, r_vec=r_vec, **kwargs_selection)
        rise=np.log10(kappa_profile[:-1])-np.log10(kappa_profile[1:])
        run=np.log10(r_vec[:-1])-np.log10(r_vec[1:])
        slope=np.append(0,rise/run)#add a zero at r=0 so that slope as same size as r_vec
        if r_eval==None:
            return slope 
        else:
            closest_r = self._find_nearest(r_vec,r_eval) #just takes closest r. Could rebuild it to interpolate.
            return slope[r_vec==closest_r]

    def effective_radius_light(self, outer_radius=10, center=None, coordinates=None,
                               initial_guess=1, initial_delta_pix=10, 
                               n_iter=10, return_model=False, return_accuracy=False,
                               circular_mask_radius=None, **kwargs_selection):
        """Computes the effective radius of the 2D surface brightness profile, 
        based on a definition similar to the half-light radius.

        Parameters
        ----------
        outer_radius : int, optional
            outer limit of integration within which half the light is calculated to estimate the effective radius, by default 10
        center : (float, float), optional
            (x, y)-coordinates of the center from which to calculate Einstein radius; if None, use the value from create_kappa_image, by default None
        coordinates : Coordinates, optional
            Instance of a Coordinates object to be used for the computation.
            If None, will use an instance based on the Instrument, by default None
        initial_guess : int, optional
            Initial guess for effective radius in arcsecond, by default 1
        initial_delta_pix : int, optional
            Initial step size in pixels before shrinking in future iterations, by default 10
        n_iter : int, optional
            Number of iterations, by default 5
        circular_mask_radius : float, optional
            If not None, multiply the flux by a circular mask with radius `circular_mask_radius` 
            to force to zero any flux outside of it.
        return_model : bool, optional
            If True, also returns the surface brightness map used to comouted the radius. By default False.
        return_accuracy : bool, optional
            if True, return a rough estimate of accuracy as well, by default False

        Returns
        -------
        float
            Effective radius

        Raises
        ------
        Warning
            If integration loop exceeds outer bound before convergence.
        """
        if kwargs_selection is None:
            kwargs_selection = {}

        light_model = ComposableLightModel(self.coolest, self.coolest_dir, **kwargs_selection)

        # select a center
        if center is None:
            center_x, center_y = light_model.estimate_center()
        else:
            center_x, center_y = center

        # get an image of the convergence
        if coordinates is None:
            x, y = self.coordinates.pixel_coordinates
        else:
            x, y = coordinates.pixel_coordinates
        # make sure to evaluate the profile such that it is centered on the image
        x_ = x + center_x
        y_ = y + center_y
        light_image = light_model.evaluate_surface_brightness(x_, y_)
        light_image[np.isnan(light_image)] = 0.
        
        #if limit of integration exceeds FoV, raise warning
        if coordinates is None:
            x_FoV = self.coolest.observation.pixels.field_of_view_x
            y_FoV = self.coolest.observation.pixels.field_of_view_y
        else:
            x_FoV = (coordinates.extent[0], coordinates.extent[1])
            y_FoV = (coordinates.extent[2], coordinates.extent[3])
        out_of_FoV=False
        if outer_radius - center_x < x_FoV[0] or outer_radius + center_x > x_FoV[1]:
            out_of_FoV=True
        if outer_radius - center_y < y_FoV[0] or outer_radius + center_y > y_FoV[1]:
            out_of_FoV=True
        if out_of_FoV is True:
            logging.warning("Outer limit of integration exceeds FoV; effective radius may not be accurate.")

        if circular_mask_radius is not None:  # circular mask that fills the FoV
            mask = (np.hypot(x, y) < circular_mask_radius).astype(float)  # 0. and 1. only
            light_image *= mask
            # import matplotlib.pyplot as plt
            # plt.imshow(np.log10(light_image))
            # plt.show()
            # raise

        r_eff, accuracy = util.effective_radius(
            light_image, x, y, outer_radius=outer_radius, initial_guess=initial_guess, 
            initial_delta_pix=initial_delta_pix, n_iter=n_iter
        )

        if return_model and return_accuracy:
            return r_eff, accuracy, light_image
        elif return_model:
            return r_eff, light_image
        elif return_accuracy:
            return r_eff, accuracy
        return r_eff

    def two_point_correlation(self, Nbins=100, rmax=None, normalize=False, 
                              use_profile_coordinates=True, coordinates=None, 
                              min_flux=None, min_flux_frac=None, 
                              return_cov=False, return_map=False,
                              **kwargs_selection):
        """
        The two point correlation function can be obtained from the covariance matrix of an image and the distances between its pixels.
        By binning the covariance matrix entries in distance (or radial) bins, one can obtain the 1D correlation function.
        There are two ways to obtain the covariance matrix:
        1) it is equivalent to the inverse Fourier transform of the power spectrum, and
        2) by calculating explicitly the covariance between any two pixels
        Here we use the first way.

        Parameters
        ----------
        Nbins : int, optional
            The number of radial bins to use for converting the 2D covariance matrix into a 1D correlation function.
        rmax : float, optional
            A value for the maximum extent of the radial bins. If none is given then it is equal to half the diagonal of the provided image.
        normalize : bool, optional
            Normalize the given image by its maximum. Default is False.
        coordinates : Coordinates, optional
            Instance of a Coordinates object to be used for the computation.
            If None, will use an instance based on the Instrument, by default None
        use_profile_coordinates : bool, optional
            If True and `coordinates=None`, uses the coordinates attached to the light profile, if available. Default is True.
        min_flux : float, optional
            Minimum flux value considered in the computation of the correlation function. 
            Default is None (i.e., no thresholding).
        min_flux_frac : float, optional
            Same as min_flux, but given as a fraction of the maximum flux value.
            If `min_flux` is not None, `min_flux_frac` is ignored.
            Default is None (i.e., no thresholding).
        return_cov : bool, optional
            If True, also returns the full covariance matrix. Default is False.
        return_cov : bool, optional
            If True, also returns the full covariance matrix. Default is False. 

        Returns
        -------
        (array, array, array)
            The location, value, and uncertainty of the 1D bins of the two-point correlation function.
            The location (radius/distance) is in the same units as the coordinates.
        """
        if kwargs_selection is None:
            kwargs_selection = {}
        if coordinates is None:
            coordinates = self.coordinates

        light_model = ComposableLightModel(self.coolest, self.coolest_dir, **kwargs_selection)

        if use_profile_coordinates is True:
            light_image, _, coordinates = light_model.surface_brightness(return_extra=True)
            if coordinates is None:
                # can be known if e.g. the underlying light profile is not pixelated
                raise ValueError("Light profile does not have any coordinates grid attached to it.")
        else:
            if coordinates is None:
                coordinates = self.coordinates
            x, y = coordinates.pixel_coordinates
            light_image = light_model.evaluate_surface_brightness(x, y)
        
        light_image = np.nan_to_num(light_image, nan=0.)
        cov_mask = np.ones_like(light_image)
        if min_flux is not None:
            logging.info(f"Setting to zero any flux below {min_flux}.")
            light_image[light_image < min_flux] = 0.
            cov_mask[light_image < min_flux] = 0.
        elif min_flux_frac is not None:
            min_flux = min_flux_frac*light_image.max()
            logging.info(f"Setting to zero any flux below {min_flux}.")
            light_image[light_image < min_flux] = 0.
            cov_mask[light_image < min_flux] = 0.
        cov_mask = cov_mask.astype(bool)

        extent = coordinates.extent
        dpix = coordinates.pixel_size
        
        if rmax is None:
            rmax = np.hypot(extent[0]-extent[1],extent[2]-extent[3])/2.0

        if normalize:
            max_image = np.amax(light_image)
            light_image = np.divide(light_image,max_image)
        
        bins, means, sdevs, cov = util.azim_averaged_two_point_correlation(
            light_image, dpix, rmax, Nbins,
        )

        if return_cov and return_map:
            return bins, means, sdevs, cov, light_image
        elif return_cov:
            return bins, means, sdevs, cov
        elif return_map:
            return bins, means, sdevs, light_image
        return bins, means, sdevs
    

    def total_magnitude(self, outer_radius=10, center=None, coordinates=None,
                        no_re_eval=False, flux_factor=None, mag_zero_point=None, **kwargs_selection):
        """Computes the effective radius of the 2D surface brightness profile, 
        based on a definition similar to the half-light radius.

        Parameters
        ----------
        outer_radius : int, optional
            outer limit of integration within which half the light is calculated 
            to integrate the flux, by default 10
        no_re_eval : bool, option
            If True, do re-evaluate the light profile (only relevant for pixelated profiles). Default is False.
        center : (float, float), optional
            (x, y)-coordinates of the center from which to calculate Einstein radius; if None, use the value from create_kappa_image, by default None
        coordinates : Coordinates, optional
            Instance of a Coordinates object to be used for the computation.
            If None, will use an instance based on the Instrument, by default None
        mag_zero_point : float, optional
            Magnitude zero-point corresponding to 1 electron per second. 
            Must be given when no mag_zero_point has been found in the self.coolest object.
        
        TODO: flux_factor is temporary, this will be removed in the future.

        Returns
        -------
        float
            Total magnitude from the flux integrated over the field of view.
        """
        if kwargs_selection is None:
            kwargs_selection = {}

        light_model = ComposableLightModel(self.coolest, self.coolest_dir, **kwargs_selection)

        # get an image of the convergence
        if no_re_eval:
            light_image, _, coordinates = light_model.surface_brightness(return_extra=True)
            light_image[np.isnan(light_image)] = 0.
            x, y = coordinates.pixel_coordinates
        else:
            # select a center
            if center is None:
                center_x, center_y = light_model.estimate_center()
            else:
                center_x, center_y = center
            if coordinates is None:
                x, y = self.coordinates.pixel_coordinates
            else:
                x, y = coordinates.pixel_coordinates
            # make sure to evaluate the profile such that it is centered on the image
            x_ = x + center_x
            y_ = y + center_y
            light_image = light_model.evaluate_surface_brightness(x_, y_)
            light_image[np.isnan(light_image)] = 0.

        # retrieve the zero-point from the
        if self.coolest.observation.mag_zero_point is not None:
            logging.info(f"Using magnitude zero-point from self.coolest object ({mag_zero_point}).")
            mag_zero_point = self.coolest.observation.mag_zero_point
        elif mag_zero_point is None:
            raise ValueError("No `mag_zero_point` has been found in the COOLEST object, "
                             "hence `mag_zero_point` must be provided.")
        else:
            logging.info(f"Using the magnitude zero-point ({mag_zero_point}.)")

        if outer_radius is not None:
            mask = (np.hypot(x, y) < outer_radius).astype(float)  # 0. and 1. only
            light_image *= mask

        # compute the magnitude
        flux_tot = light_image.sum()
        if flux_factor is not None:
            flux_tot *= flux_factor  # temporary feature
        mag_tot = -2.5*np.log10(flux_tot) + mag_zero_point
    
        return mag_tot

    def ellipticity_from_moments(self, center=None, coordinates=None, **kwargs_selection):
        """Estimates the axis ratio and position angle of the model map 
        based on central moments of the image.

        Parameters
        ----------
        center : (float, float), optional
            (x, y)-coordinates of the center from which to calculate Einstein radius; if None, use the value from create_kappa_image, by default None
        coordinates : Coordinates, optional
            Instance of a Coordinates object to be used for the computation.
            If None, will use an instance based on the Instrument, by default None
        
        Returns
        -------
        float
            Ellipticity measurement (axis)

        Raises
        ------
        Warning
            If integration loop exceeds outer bound before convergence.
        """
        if kwargs_selection is None:
            kwargs_selection = {}

        light_model = ComposableLightModel(self.coolest, self.coolest_dir, **kwargs_selection)

        # select a center
        if center is None:
            center_x, center_y = light_model.estimate_center()
        else:
            center_x, center_y = center

        # get an image of the convergence
        if coordinates is None:
            x, y = self.coordinates.pixel_coordinates
            pixel_size = self.coordinates.pixel_size
        else:
            x, y = coordinates.pixel_coordinates
            pixel_size = coordinates.pixel_size
            
        # make sure to evaluate the profile such that it is centered on the image
        x_ = x + center_x
        y_ = y + center_y
        light_image = light_model.evaluate_surface_brightness(x_, y_)
        light_image[np.isnan(light_image)] = 0.

        # compute the ellipticity and position angle
        phi_rad, q = util.ellipticity_from_moments(light_image, pixel_size)
        phi = phi_rad * 180./np.pi + 90. # conversion to COOLEST conventions

        return q, phi

    def lensing_information(self, a=16, b=0, 
                            noise_map=None, arc_mask=None, theta_E=None,
                            entity_idx_theta_E=0, profile_idx_theta_E=0):
        """
        Computes the 'lensing information' defined in Yi Tan et al. 2023, Equations (8) and (9).
        https://ui.adsabs.harvard.edu/abs/2023arXiv231109307T/abstract
        """
        data = self.coolest.observation.pixels.get_pixels()
        # TODO: subtract the lens light
        print("WARNING: no lens light subtracted; assuming the data contains only the arcs.")
        if noise_map is None:
            raise NotImplementedError("Getting the noise map from the COOLEST instance is yet implemented.")
        if theta_E is None:
            mass_profile = self.coolest.lensing_entities[entity_idx_theta_E].mass_model[profile_idx_theta_E]
            theta_E = mass_profile.parameters['theta_E'].point_estimate.value
        x, y = self.coordinates.pixel_coordinates
        I, theta_E, phi_ref, mask = util.lensing_information(
            data, x, y, theta_E, noise_map, a=a, b=b, arc_mask=arc_mask
        )
        return I, theta_E, phi_ref, mask
    
    @staticmethod
    def _find_nearest(array, value):
        """subfunction to find nearest closest element in array to value"""
        array = np.asarray(array)
        idx = (np.abs(array - value)).argmin()
        return array[idx]
    


import os
import json
import jsonpickle
import math

from coolest.template.standard import COOLEST
from coolest.template.lazy import *
from coolest.template.classes.parameter import PointEstimate, PosteriorStatistics, Prior
from coolest.template.info import all_supported_choices as support


__all__ = ['JSONSerializer']


class JSONSerializer(object):
    """Object that write a COOLEST object to a JSON template file,
    or loads the content of a JSON template file in the COOLEST format.

    For reading, either a plain JSON format or the one generated via `jsonpickle` can be provided.
    For the latter, the JSON file should end with the suffix `'_pyAPI'` (before the .json extension).

    NOTE: the support for reading the template with `jsonpickle` will probably 
    be suppressed in the future. Hence we advise users to work with the 
    pure JSON template files instead.
    
    Parameters
    ----------
    file_path_no_ext : str
        Path to the JSON template, or the one to be created.
        It should NOT include the .json extension nor the optional '_pyAPI' suffix
    obj : object, optional
        Instance of the COOLEST object (from the `standard` module) 
        to be encoded, by default None
    indent : int, optional
        Number of spaces used to indent lines in the JSON file, by default 2
    check_external_files : bool, optional
        If True, will check the existence of external (e.g., FITS files) 
        specified within the JSON template, by default True

    Raises
    ------
    ValueError
        If the provided path to the JSON file is not an absolute path
    ValueError
        If the provided path contains the .json extension
    """

    # suffix to filename to distinguish files that can be read using jsonpickle
    _api_suffix = '_pyAPI'

    def __init__(self,
                 file_path_no_ext: str, 
                 obj: object = None, 
                 indent: int = 2,
                 check_external_files: bool = True) -> None:
        if not os.path.isabs(file_path_no_ext):
            raise ValueError("Path to JSON file must be an absolute path")
        if file_path_no_ext[-5:].lower() == '.json':
            raise ValueError("The provided template name should not contain the JSON extension")
        self.path = file_path_no_ext
        self._json_dir = os.path.dirname(file_path_no_ext)
        self.obj = obj
        self.indent = indent
        self._check_files = check_external_files

    def dump_simple(self, exclude_keys=None):
        """Write to disk the template file, in plain JSON format.

        Parameters
        ----------
        exclude_keys : bool, optional
            List of class attributes that should not be included 
            in the JSON file, by default None
        """
        if exclude_keys is None:
            exclude_keys = self.obj.exclude_keys
        json_path = self.path + '.json'
        with open(json_path, 'w') as f:
            f.write(self.obj.to_JSON(indent=self.indent, exclude_keys=exclude_keys))

    def dump_jsonpickle(self):
        """Write to disk the template file, using the `jsonpickle` package

        WARNING: this feature may be dropped in the future.
        """
        json_path = self.path + self._api_suffix + '.json'
        result = jsonpickle.encode(self.obj, indent=self.indent)
        with open(json_path, 'w') as f:
            f.write(result)

    def load(self, skip_jsonpickle=False, verbose=True):
        """Read the JSON template file and build up the corresponding COOLEST object.
        It will first try to load the '_pyAPI' template if it exists using `jsonpickle`, 
        otherwise it will fall back to reading the pure json template.

        Parameters
        ----------
        skip_jsonpickle : bool, optional
            If True, will not try to read the _pyAPI template with jsonpickle first, by default False
        verbose : bool, optional
            If True, prints useful output for debugging, by default False

        Returns
        -------
        COOLEST object
            COOLEST object that corresponds to the JSON template
        """
        json_path = self.path + '.json'
        jsonpickle_path = self.path + self._api_suffix + '.json'
        if os.path.exists(jsonpickle_path) and not skip_jsonpickle:
            instance = self.load_jsonpickle(jsonpickle_path)
        else:
            if verbose:
                print(f"Template file '{jsonpickle_path}' not found, now trying to read '{json_path}'.")
            instance = self.load_simple(json_path, as_object=True)
        assert isinstance(instance, COOLEST)
        return instance

    def load_simple(self, json_path, as_object=True):
        """Read the JSON template file and build up the corresponding COOLEST object. 

        Parameters
        ----------
        json_path: str
            Path to the json file to be read.
        as_object : bool, optional
            _description_, by default True

        Returns
        -------
        COOLEST object
            COOLEST object that corresponds to the JSON template
        """
        with open(json_path, 'r') as f:
            content = json.loads(f.read())
        if not as_object:
            return content  # dictionary
        return self._json_to_coolest(content)  # COOLEST object

    def load_jsonpickle(self, jsonpickle_path):
        """Read the JSON template file and build up the corresponding COOLEST object
        using the `jsonpickle`.

        Parameters
        ----------
        jsonpickle_path: str
            Path to the json file to be read with `jsonpickle`.

        Returns
        -------
        COOLEST object
            COOLEST object that corresponds to the JSON template
        """
        with open(jsonpickle_path, 'r') as f:
            content = jsonpickle.decode(f.read())
        return content  # COOLEST object

    def _json_to_coolest(self, json_content):
        """Creates from scratch a COOLEST instance based on the content of a JSON
        file, given as a nested dictionnary.

        Parameters
        ----------
        json_content : dict
            Content of the JSON template file

        Returns
        -------
        COOLEST object
            COOLEST object that corresponds to the JSON template
        """
        c = json_content  # shorter

        # MODE
        mode = self._check_mode(c['mode'])

        # LENSING ENTITIES {GALAXY, MASSFIELDS}
        lensing_entities = self._setup_lensing_entities(c['lensing_entities'])

        # COSMOLOGY
        cosmology = self._setup_cosmology(c['cosmology'])

        # COORDINATES
        coordinates_origin = self._setup_coordinates(c['coordinates_origin'])

        # OBSERVATION
        observation = self._setup_observation(c['observation'])

        # INSTRUMENT
        instrument = self._setup_instrument(c['instrument'])

        # METADATA
        metadata = self._check_metadata(c['meta'])

        # instantiate the master class
        coolest = COOLEST(mode,
                          coordinates_origin,
                          lensing_entities,
                          observation,
                          instrument,
                          cosmology=cosmology,
                          metadata=metadata)

        # check consistency across the whole coolest object
        self._validate_global(coolest)
        return coolest

    @staticmethod
    def _validate_global(coolest):
        """Performs consistency checks regarding some key properties of the COOLEST object.
        For instance, it checks that the pixel size of both the observation and
        the instrument are consistent.
        The checks performed here are those that cannot be handled by individual 
        class constructors called during instantiation of the COOLEST object.

        Parameters
        ----------
        coolest : COOLEST object
            Instance of a COOLEST object

        Raises
        ------
        ValueError
            In case observed instrumental pixel sizes are inconsistent
        """
        # PIXEL SIZE
        instru_pix_size = coolest.instrument.pixel_size
        obs_pix_size = coolest.observation.pixels.pixel_size
        isclose_bool = math.isclose(instru_pix_size, obs_pix_size,
                                    rel_tol=1e-09, abs_tol=0.0)
        if obs_pix_size not in (0, None) and not isclose_bool:
            raise ValueError(f"Pixel size of observation ({obs_pix_size}) is inconsistent with "
                             f"the instrument pixel size ({instru_pix_size})")

        # TODO: add extra checks

    def _setup_instrument(self, instru_in):
        psf_settings = instru_in.pop('psf')
        psf = self._setup_psf(psf_settings)
        instru_out = Instrument(psf=psf, **instru_in)
        return instru_out

    def _setup_observation(self, obs_in):
        pixels_settings = obs_in.pop('pixels')
        pixels = self._setup_grid(pixels_settings, PixelatedRegularGrid)
        noise_settings = obs_in.pop('noise')
        noise = self._setup_noise(noise_settings)
        obs_out = Observation(pixels=pixels, noise=noise, **obs_in)
        return obs_out

    def _setup_coordinates(self, coord_orig_in):
        return CoordinatesOrigin(**coord_orig_in)

    def _setup_cosmology(self, cosmology_in):
        return Cosmology(**cosmology_in)
    
    def _setup_lensing_entities(self, entities_in):
        entities_out = []
        for entity_in in entities_in:
            entities_out.append(self._setup_lensing_entity(entity_in))
        return LensingEntityList(*entities_out)
    
    def _setup_lensing_entity(self, entity_in):
        if entity_in['type'] == 'Galaxy':
            entity_out = Galaxy(entity_in['name'],
                                entity_in['redshift'],
                                light_model=self._setup_model(entity_in, 'light_model'),
                                mass_model=self._setup_model(entity_in, 'mass_model'))
        elif entity_in['type'] == 'MassField':
            entity_out = MassField(entity_in['name'],
                                   entity_in['redshift'],
                                   mass_model=self._setup_model(entity_in, 'mass_model'))
        else:
            raise ValueError(f"Supported lensing entities are "
                             f"{support['lensing_entities']}")
        self._update_parameters(entity_in, entity_out)
        return entity_out
    
    def _setup_model(self, entity_in, model_type):
        profile_types = [profile['type'] for profile in entity_in[model_type]]
        if model_type == 'light_model':
            model_out = LightModel(*profile_types)
        elif model_type == 'mass_model':
            model_out = MassModel(*profile_types)
        else:
            raise ValueError("The `model_type` can only be 'light_model' or 'mass_model'")
        return model_out
    
    def _update_parameters(self, entity_in, entity_out):
        self._update_parameters_values(entity_in, entity_out, 'mass_model')
        self._update_parameters_values(entity_in, entity_out, 'light_model')

    def _update_parameters_values(self, entity_in, entity_out, model_type):
        if model_type not in entity_in:
            return
        for i, profile in enumerate(entity_in[model_type]):
            # get the corresponding profile object in the model being updated
            profile_out = getattr(entity_out, model_type)[i]

            for name, values in profile['parameters'].items():

                # pop the id as this was already set at instantiation
                _ = values.pop('id', None)

                # pixelated profiles, for now only one value given (point estimate)
                if 'Grid' in profile['type']:
                    self._update_grid_parameter(profile_out, name, values)

                # other (analytical) parameters
                else:
                    self._update_std_parameter(profile_out, name, values)

    def _update_grid_parameter(self, profile_out, name, values):
        if name != 'pixels':
            raise NotImplementedError("Support for grid parameters other than "
                                      "'pixels' is not implemented.")
        if 'Regular' in profile_out.type:
            pixels = self._setup_grid(values, PixelatedRegularGrid)
            profile_out.parameters['pixels'] = pixels
        elif 'Irregular' in profile_out.type:
            pixels = self._setup_grid(values, IrregularGrid)
            profile_out.parameters['pixels'] = pixels
        else:
            raise ValueError(f"Unknown grid profile ({profile_out.type})")
        
    def _update_std_parameter(self, profile_out, name, values):
        pt_estim = PointEstimate(**values['point_estimate'])
        profile_out.parameters[name].set_point_estimate(pt_estim)
        if 'posterior_stats' in values:
            post_stats = PosteriorStatistics(**values['posterior_stats'])
            profile_out.parameters[name].set_posterior(post_stats)
        if 'prior' in values:
            prior = self._setup_prior(values['prior'])
            profile_out.parameters[name].set_prior(prior)

    def _setup_prior(self, prior_in):
        from coolest.template.classes import probabilities as proba_module
        prior_type = prior_in.pop('type')
        if prior_type is None:
            return Prior()
        PriorClass = getattr(proba_module, prior_type)
        return PriorClass(**prior_in)

    def _setup_noise(self, noise_in):
        from coolest.template.classes import noise as noise_module
        noise_type = noise_in.pop('type')
        if noise_type is None:
            return Noise()
        if noise_type not in support['noise_types']:
            raise ValueError(f"Noise type must be in {support['noise_types']}")
        NoiseClass = getattr(noise_module, noise_type)
        if noise_type == 'NoiseMap':
            pixels_settings = noise_in.pop('noise_map')
            noise_map = self._setup_grid(pixels_settings, PixelatedRegularGrid)
            noise_out = NoiseClass(noise_map=noise_map, **noise_in)
        elif noise_type == 'NoiseRealization':
            pixels_settings = noise_in.pop('noise_realization')
            noise_realization = self._setup_grid(pixels_settings, PixelatedRegularGrid)
            noise_out = NoiseClass(noise_realization=noise_realization, **noise_in)
        elif noise_type == 'DrizzledNoise':
            pixels_settings = noise_in.pop('wht_map')
            wht_map = self._setup_grid(pixels_settings, PixelatedRegularGrid)
            noise_out = NoiseClass(wht_map=wht_map, **noise_in)
        else:
            noise_out = NoiseClass(**noise_in)
        return noise_out

    def _setup_psf(self, psf_in):
        from coolest.template.classes import psf as psf_module
        psf_type = psf_in.pop('type')
        if psf_type is None:
            return PSF()
        if psf_type not in support['psf_types']:
            raise ValueError(f"PSF type must be in {support['psf_types']}")
        PSFClass = getattr(psf_module, psf_type)        
        if psf_type == 'PixelatedPSF':
            pixels_settings = psf_in.pop('pixels')
            pixels = self._setup_grid(pixels_settings, PixelatedRegularGrid)
            psf_out = PSFClass(pixels=pixels, **psf_in)
        else:
            psf_out = PSFClass(**psf_in)
        return psf_out

    def _setup_grid(self, grid_in, GridClass):
        fits_path = grid_in.pop('fits_file')['path']
        if fits_path is not None and os.path.isabs(fits_path):
            raise ValueError(f"FITS file '{fits_path}' must be a relative path instead, "
                             f"and placed in the same directory as the JSON file")
        return GridClass(fits_path, 
            check_fits_file=self._check_files, 
            fits_file_dir=self._json_dir,
            **grid_in)

    def _check_mode(self, mode_in):
        mode_out = str(mode_in)
        if mode_out not in support['modes']:
            raise ValueError(f"The template mode can only be in '{SUPPORTED_MODES}'")
        return mode_out

    def _check_metadata(self, meta_in):
        meta_out = meta_in  # TODO: might do more checks here
        return meta_out



__author__ = 'aymgal'

import numpy as np
from typing import Dict

from coolest.template.classes.base import APIBaseObject
from coolest.template.classes.parameter import (Parameter, 
                                                LinearParameter, LinearParameterSet,
                                                NonLinearParameter, HyperParameter)


__all__ = ['Profile' ,'AnalyticalProfile']


class Profile(APIBaseObject):
    """Abstract class for any light or mass profile.

    Parameters
    ----------
    parameters : Dict[(str, Parameter)]
        Dictionary of Parameters objects keyed by parameter name.
    """

    def __init__(self, parameters: Dict[(str, Parameter)]) -> None:
        self.type = self.__class__.__name__  # name of children class
        self.parameters = parameters
        self.id = None
        super().__init__()


class AnalyticalProfile(Profile):
    """Abstract class for an light or mass profile described 
    by one or several analytical functions.

    Parameters
    ----------
    parameters : Dict[(str, Parameter)]
        Dictionary of Parameters objects keyed by parameter name.
    """

    def __init__(self, parameters: Dict[(str, Parameter)]) -> None:
        """"""
        super().__init__(parameters)
        
    def total_num_params(self, include_fixed=False, include_hyper=True):
        """Compute the number of parameter of the profile.

        Parameters
        ----------
        include_fixed : bool, optional
            Include fixed parameters in the count, by default False
        include_hyper : bool, optional
            Include the hyper-parameters in the count, by default True

        Returns
        -------
        int
            Number of parameters
        """
        count = 0
        for name, parameter in self.parameters.items():
            if isinstance(parameter, (NonLinearParameter, LinearParameter)):
                if not parameter.fixed or include_fixed:
                    count += 1
            elif isinstance(parameter, LinearParameterSet):
                if not parameter.fixed or include_fixed:
                    count += parameter.num_values
            elif isinstance(parameter, HyperParameter) and include_hyper:
                if not parameter.fixed or include_fixed:
                    count += 1
        return count



__author__ = 'aymgal'

from coolest.template.classes.profile import Profile, AnalyticalProfile
from coolest.template.classes.parameter import NonLinearParameter
from coolest.template.classes.parameter import (DefinitionRange,
                                            PixelatedRegularGridParameter)
from coolest.template.classes.grid import PixelatedRegularGrid


__all__ = [
    'SIE', 
    'NIE',
    'PEMD',
    'SPEMD',
    'NFW',
    'Chameleon',
    'ExternalShear', 
    'ConvergenceSheet', 
    'PixelatedRegularGridPotential',
]
SUPPORTED_CHOICES = __all__


class SIE(AnalyticalProfile):
    """Singular isothermal ellipsoidal profile.

    This profile is described by the following parameters:

    - 'theta_E': Einstein radius
    - 'q': axis ratio (semi-major axis / semi-minor axis)
    - 'phi': position angle
    - 'center_x': position along the x coordinate
    - 'center_y': position along the y coordinate

    """
    
    def __init__(self):
        parameters = {
            'theta_E': NonLinearParameter("Einstein radius",
                                          DefinitionRange(min_value=0.0),
                                          latex_str=r"$\theta_{\rm E}$"),
            'q': NonLinearParameter("Axis ratio, semi-minor axis / semi-major axis",
                                     DefinitionRange(min_value=0.0, max_value=1.0),
                                     latex_str=r"$q$"),
            'phi': NonLinearParameter("Position angle",
                                     DefinitionRange(min_value=-90., max_value=90.),
                                     latex_str=r"$\phi$"),
            'center_x': NonLinearParameter("Profile center along x coordinates",
                                           DefinitionRange(),
                                           latex_str=r"$x_0$"),
            'center_y': NonLinearParameter("Profile center along y coordinates",
                                           DefinitionRange(),
                                           latex_str=r"$y_0$")
        }
        super().__init__(parameters)


class NIE(AnalyticalProfile):
    """Non-singular isothermal ellipsoidal profile.

    This profile is described by the following parameters:

    - 'theta_E': Einstein radius
    - 'r_core: core radius
    - 'q': axis ratio (semi-major axis / semi-minor axis)
    - 'phi': position angle
    - 'center_x': position along the x coordinate
    - 'center_y': position along the y coordinate

    """
    
    def __init__(self):
        parameters = {
            'theta_E': NonLinearParameter("Einstein radius",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$\theta_{\rm E}$"),
            'r_core': NonLinearParameter("Core radius",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$r_{\rm c}$"),
            'q': NonLinearParameter("Axis ratio, semi-minor axis / semi-major axis",
                                     DefinitionRange(min_value=0.0, max_value=1.0),
                                     latex_str=r"$q$"),
            'phi': NonLinearParameter("Position angle",
                                     DefinitionRange(min_value=-90., max_value=90.),
                                     latex_str=r"$\phi$"),
            'center_x': NonLinearParameter("Profile center along x coordinates",
                               DefinitionRange(),
                               latex_str=r"$x_0$"),
            'center_y': NonLinearParameter("Profile center along y coordinates",
                               DefinitionRange(),
                               latex_str=r"$y_0$")
        }
        super().__init__(parameters)


class PEMD(AnalyticalProfile):
    """Power-law elliptical mass density profile, 
    also called the Elliptical Power-law profile (EPL).

    This profile is described by the following parameters:

    - 'gamma': logarithmic radial slope
    - 'theta_E': Einstein radius
    - 'q': axis ratio (semi-major axis / semi-minor axis)
    - 'phi': position angle
    - 'center_x': position along the x coordinate
    - 'center_y': position along the y coordinate
    """
    
    def __init__(self):
        parameters = {
            'gamma': NonLinearParameter("Mass density slope at Einstein radius",
                               DefinitionRange(min_value=1.0, max_value=3.0),
                               latex_str=r"$\gamma$"),
            'theta_E': NonLinearParameter("Einstein radius",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$\theta_{\rm E}$"),
            'q': NonLinearParameter("Axis ratio, semi-minor axis / semi-major axis",
                                     DefinitionRange(min_value=0.0, max_value=1.0),
                                     latex_str=r"$q$"),
            'phi': NonLinearParameter("Position angle",
                                     DefinitionRange(min_value=-90., max_value=90.),
                                     latex_str=r"$\phi$"),
            'center_x': NonLinearParameter("Profile center along x coordinates",
                               DefinitionRange(),
                               latex_str=r"$x_0$"),
            'center_y': NonLinearParameter("Profile center along y coordinates",
                               DefinitionRange(),
                               latex_str=r"$y_0$")
        }
        super().__init__(parameters)


class SPEMD(AnalyticalProfile):
    """Softened power-law elliptical mass density profile.

    This profile is described by the following parameters:

    - 'gamma': negative logarithmic radial slope
    - 'theta_E': Einstein radius
    - 'r_core': core radius
    - 'q': axis ratio (semi-major axis / semi-minor axis)
    - 'phi': position angle
    - 'center_x': position along the x coordinate
    - 'center_y': position along the y coordinate
    """
    
    def __init__(self):
        parameters = {
            'gamma': NonLinearParameter("Mass density slope at Einstein radius",
                               DefinitionRange(min_value=1.0, max_value=3.0),
                               latex_str=r"$\gamma$"),
            'theta_E': NonLinearParameter("Einstein radius",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$\theta_{\rm E}$"),
            'r_core': NonLinearParameter("Core radius",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$r_{\rm c}$"),
            'q': NonLinearParameter("Axis ratio, semi-minor axis / semi-major axis",
                                     DefinitionRange(min_value=0.0, max_value=1.0),
                                     latex_str=r"$q$"),
            'phi': NonLinearParameter("Position angle",
                                     DefinitionRange(min_value=-90., max_value=90.),
                                     latex_str=r"$\phi$"),
            'center_x': NonLinearParameter("Profile center along x coordinates",
                               DefinitionRange(),
                               latex_str=r"$x_0$"),
            'center_y': NonLinearParameter("Profile center along y coordinates",
                               DefinitionRange(),
                               latex_str=r"$y_0$")
        }
        super().__init__(parameters)


class NFW(AnalyticalProfile):
    """Navarro-Frenk-White mass profile.

    This profile is described by the following parameters:

    - 'r_s': transition radis from logarithmic radial slope -1 to -3 
    - 'rho_c': characteristic mass density
    - 'q': axis ratio (semi-major axis / semi-minor axis)
    - 'phi': position angle
    - 'center_x': position along the x coordinate
    - 'center_y': position along the y coordinate
    """
    
    def __init__(self):
        parameters = {
            'r_s': NonLinearParameter("Radius at which the mass density profile slope transitions from -1 to -3",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$r_{\rm s}$"),
            'rho_c': NonLinearParameter("Characteristic mass density",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$\rho_c$"),
            'q': NonLinearParameter("Axis ratio, semi-minor axis / semi-major axis",
                                     DefinitionRange(min_value=0.0, max_value=1.0),
                                     latex_str=r"$q$"),
            'phi': NonLinearParameter("Position angle",
                                     DefinitionRange(min_value=-90., max_value=90.),
                                     latex_str=r"$\phi$"),
            'center_x': NonLinearParameter("Profile center along x coordinates",
                               DefinitionRange(),
                               latex_str=r"$x_0$"),
            'center_y': NonLinearParameter("Profile center along y coordinates",
                               DefinitionRange(),
                               latex_str=r"$y_0$")
        }
        super().__init__(parameters)


class Chameleon(AnalyticalProfile):
    """'Chameleon' mass profile defined as the difference between 
    two concentric NIE profiles.

    This profile is described by the following parameters:
    
    - 'b': normalization of the profile
    - 's_c': core radius of the innver NIE
    - 's_t': core radius of the outer NIE
    - 'q': axis ratio (semi-major axis / semi-minor axis)
    - 'phi': position angle
    - 'center_x': position along the x coordinate
    - 'center_y': position along the y coordinate
    """
    
    def __init__(self):
        parameters = {
            'b': NonLinearParameter("Normalization of the profile",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$\alpha_1$"),
            's_c': NonLinearParameter("Core radius of inner NIE",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$w{\rm c}$"),
            's_t': NonLinearParameter("Core radius of outer NIE",
                               DefinitionRange(min_value=0.0),
                               latex_str=r"$w{\rm t}$"),
            'q': NonLinearParameter("Axis ratio, semi-minor axis / semi-major axis",
                                     DefinitionRange(min_value=0.0, max_value=1.0),
                                     latex_str=r"$q$"),
            'phi': NonLinearParameter("Position angle",
                                     DefinitionRange(min_value=-90., max_value=90.),
                                     latex_str=r"$\phi$"),
            'center_x': NonLinearParameter("Profile center along x coordinates",
                               DefinitionRange(),
                               latex_str=r"$x_0$"),
            'center_y': NonLinearParameter("Profile center along y coordinates",
                               DefinitionRange(),
                               latex_str=r"$y_0$")
        }
        super().__init__(parameters)


class ExternalShear(AnalyticalProfile):
    """External shear defined with a strength and orientation.
    The 'origin' of the external shear is by convention fixed to coordinates (0, 0).
    
    This profile is described by the following parameters:

    - 'gamma_ext': strength of the shear field
    - 'phi_ext': orientation of the shear field
    """
    
    def __init__(self):
        parameters = {
            'gamma_ext': NonLinearParameter("Strength of external shear",
                               DefinitionRange(min_value=0., max_value=1.0),
                               latex_str=r"$\gamma_{\rm ext}$"),
            'phi_ext': NonLinearParameter("Orientation of external shear",
                               DefinitionRange(min_value=-90., max_value=90.),
                               latex_str=r"$\phi_{\rm ext}$"),
        }
        super().__init__(parameters)


class ConvergenceSheet(AnalyticalProfile):
    """Convergence 'sheet', infinite and uniform mass density profile.
    The 'origin' of the convergence sheet is by convention fixed to coordinates (0, 0).
    The sheet is allowed to be negative, e.g. can describe an arbitrary 
    mass-sheet transformation.
    
    This profile is described by the following parameters:

    - 'kappa_s': convergence of the uniform mass density sheet
    """
    
    def __init__(self):
        parameters = {
            'kappa_s': NonLinearParameter(
                "Convergence value of the uniform mass density sheet",
                DefinitionRange(min_value=-1e8, max_value=1e8),
                latex_str=r"$\kappa_{\rm s}$"
            ),
        }
        super().__init__(parameters)


class PixelatedRegularGridPotential(Profile):
    """Lens potential defined on a grid of regular pixels.

    This profile is described by the following parameters:

    - 'pixels': 2D array of pixel values
    """

    def __init__(self):
        parameters = {
            'pixels': PixelatedRegularGridParameter("Pixel values")
        }
        super().__init__(parameters)



__author__ = 'aymgal'


from coolest.template.classes.base import APIBaseObject
from coolest.template.classes.observation import Observation
from coolest.template.classes.instrument import Instrument
from coolest.template.classes.lensing_entity_list import LensingEntityList
from coolest.template.classes.regularization_list import RegularizationList
from coolest.template.classes.likelihood_list import LikelihoodList
from coolest.template.classes.coordinates import CoordinatesOrigin
from coolest.template.classes.cosmology import Cosmology


__all__ = ['COOLEST']


SUPPORTED_MODES = ['MOCK', 'MAP', 'DOC']

MOCK_EXCLUDE_KEYS = [
    'documentation',
    'latex_str',
    'units',
    'fixed',
    'posterior_stats',
    'definition_range',
    'likelihoods',
    'regularizations',
]

MAP_EXCLUDE_KEYS = [
    'documentation',
    'latex_str',
    'units',
    'definition_range',
]

DOC_EXCLUDE_KEYS = [
    'posterior_stats',
    'prior',
    'fixed',
    'id',
]


class COOLEST(APIBaseObject):
    """Main component of COOLEST. This object stores all relevant quantities
    to a given lens model, as a hierarchy of python objects whose attributes 
    describe the Observation, Instrument, Coordinates, Lens models, etc.

    Parameters
    ----------
    mode : str
        Mode of the stored lens model described in the COOLEST object.
        Should be one of the following:

        - `'MOCK'`: the object describes a mock observation of a lens;

        - `'MAP'`: the object describes a maximum a posteriori point estimates 
        (best-fit values) lens model, possibly including statistics 
        regarding the posterior distribution of lens model parameters;

        - `'DOC'`: the object contains all description of profiles and parameters,
        (only relevant for generating documentation products).

        The mode affects which attributes a stored in the JSON representation 
        of the COOLEST object.
        
    coordinates_origin : CoordinatesOrigin
        Instance of a CoordinatesOrigin object.
    lensing_entities : LensingEntityList
        Instance of a LensingEntityList object.
    observation : Observation
        Instance of a Observation object.
    instrument : Instrument
        Instance of a Instrument object.
    cosmology : Cosmology, optional
        Instance of a Cosmology object, by default None
    metadata : dict, optional
        Any additional data to be stored in the COOLEST format, by default None

    Raises
    ------
    ValueError
        If the provided mode is not supported.
    """

    def __init__(self,
                 mode: str,
                 coordinates_origin: CoordinatesOrigin,
                 lensing_entities: LensingEntityList,
                 observation: Observation,
                 instrument: Instrument,
                 cosmology: Cosmology = None,
                 # likelihoods: LikelihoodList = None,
                 # regularizations: RegularizationList = None,
                 metadata: dict = None):
        if mode not in SUPPORTED_MODES:
            raise ValueError(f"COOLEST mode '{mode}' must be in {SUPPORTED_MODES}")
        self.mode = mode
        if coordinates_origin is None:
            coordinates_origin = CoordinatesOrigin('0', '0')
        self.coordinates_origin = coordinates_origin
        self.lensing_entities = lensing_entities
        self.observation = observation
        self.instrument  = instrument
        self.cosmology   = cosmology
        # self.likelihoods = likelihoods
        # self.regularizations = regularizations

        self.standard = 'COOLEST'
        if metadata is None:
            metadata = {}
        self.meta = metadata

    @property
    def exclude_keys(self):
        if self.mode == 'MOCK':
            return MOCK_EXCLUDE_KEYS
        elif self.mode == 'MAP':
            return MAP_EXCLUDE_KEYS
        elif self.mode == 'DOC':
            return DOC_EXCLUDE_KEYS
        else:
            return []



__author__ = 'aymgal'


import numpy as np
from scipy import special

from coolest.template.classes.profiles.mass import (PEMD as TemplatePEMD,
                                                    ExternalShear as TemplateExternalShear,
                                                    ConvergenceSheet as TemplateConvergenceSheet)
from coolest.api.profiles import util



class BaseMassProfile(object):
    """Base class to define a mass profile to compute lensing quantities.
    Each specific class must be consistent with the equivalent class from the
    coolest.template submodule.

    NOTE: in the future, a new coolest.profiles submodule will merge
    profile definitions that are currently split between coolest.template and coolest.api. 
    """

    _template_class = None

    def potential(self, **params):
        raise NotImplementedError(f"The method potential() is not defined "
                                  f"for profile '{self.__class__.__name__}'")

    def deflection(self, **params):
        raise NotImplementedError(f"The method deflection() is not defined "
                                  f"for profile '{self.__class__.__name__}'")

    def convergence(self, **params):
        raise NotImplementedError(f"The method convergence() is not defined "
                                  f"for profile '{self.__class__.__name__}'")

    def hessian(self, **params):
        raise NotImplementedError(f"The method hessian() is not defined "
                                  f"for profile '{self.__class__.__name__}'")

    @property
    def template_class(self):
        if self._template_class is None:
            raise RuntimeError("No template class has been set by mass profile class")
        return self._template_class

    @property
    def type(self):
        return self.template_class.type

    @property
    def parameter_names(self):
        return list(self.template_class.parameters.keys())


class PEMD(BaseMassProfile):

    """
    Power-law Elliptical Mass Distribution (a.k.a. Elliptical Power-law)
    This follows implementations in lenstronomy (:cite:t:`lenstronomy2018`:, :cite:t:`lenstronomy2021`:) based on the formulae :cite:p:`Tessore2015`:.
    """

    # TODO: use parameter values (point estimates, prior, etc...) contained in the template?
    _template_class = TemplatePEMD()

    def param_conv(self, theta_E, q, gamma):
        theta_E_conv = theta_E / (np.sqrt((1. + q**2) / (2. * q)))
        b = theta_E_conv * np.sqrt((1. + q**2) / 2.)
        t = gamma - 1.
        return b, t

    def potential(self, x, y, theta_E=1., gamma=2., phi=0., q=1., center_x=0., center_y=0.):
        b, t = self.param_conv(theta_E, q, gamma)
        # shift and rotate
        phi_ = util.eastofnorth2normalradians(phi)
        x_, y_ = util.shift(x, y, center_x, center_y)
        x_, y_ = util.rotate(x_, y_, phi_)
        # deflection angle
        a_x_, a_y_ = self._defl_major_axis(x_, y_, b, t, q)
        # potential
        return (x_ * a_x_ + y_ * a_y_) / (2 - t)

    def deflection(self, x, y, theta_E=1., gamma=2., phi=0., q=1., center_x=0., center_y=0.):
        b, t = self.param_conv(theta_E, q, gamma)
        # shift and rotate
        phi_ = util.eastofnorth2normalradians(phi)
        x_, y_ = util.shift(x, y, center_x, center_y)
        x_, y_ = util.rotate(x_, y_, phi_)
        # deflection angle
        a_x_, a_y_ = self._defl_major_axis(x_, y_, b, t, q)
        # rotate back
        a_x, a_y = util.rotate(a_x_, a_y_, - phi_)
        return a_x, a_y

    @staticmethod
    def _defl_major_axis(x_, y_, b, t, q):
        # evaluate the profile following to Tessore et al. 2015
        Z = np.empty(np.shape(x_), dtype=complex)
        Z.real = q * x_
        Z.imag = y_
        R = np.abs(Z)
        R = np.maximum(R, 1e-9)
        R_omega = Z * special.hyp2f1(1, t/2, 2-t/2, -(1-q)/(1+q)*(Z/Z.conj()))
        alpha = 2. / (1+q) * (b/R)**t * R_omega
        a_x_ = np.nan_to_num(alpha.real, neginf=-1e10, posinf=1e10)
        a_y_ = np.nan_to_num(alpha.imag, neginf=-1e10, posinf=1e10)
        return a_x_, a_y_

    def convergence(self, x, y, theta_E=1., gamma=2., phi=0., q=1., center_x=0., center_y=0.):
        """Returns the convergence (kappa) at the given position (x, y)"""
        phi_ = util.eastofnorth2normalradians(phi)
        x_t, y_t = util.shift_rotate_elliptical(x, y, phi_, q, center_x, center_y)
        return (3.-gamma)/2. * (theta_E / np.sqrt(x_t**2+y_t**2)) ** (gamma-1.)

    @staticmethod
    def _conv_major_axis(x_, y_, b, t, q):
        R = np.hypot(q*x_, y_)
        R = np.maximum(R, 1e-9)
        return (2 - t)/2. * (b/R)**t

    def hessian(self, x, y, theta_E=1., gamma=2., phi=0., q=1., center_x=0., center_y=0.):
        b, t = self.param_conv(theta_E, q, gamma)
        phi_ = util.eastofnorth2normalradians(phi)
        x_, y_ = util.shift(x, y, center_x, center_y)
        x_, y_ = util.rotate(x_, y_, phi_)

        # convergence
        kappa_ = self._conv_major_axis(x_, y_, b, t, q)
        kappa_ = np.nan_to_num(kappa_, neginf=-1e10, posinf=1e10)

        # deflection
        alpha_x_, alpha_y_ = self._defl_major_axis(x_, y_, b, t, q)

        #R = np.hypot(q*x, y)
        #R = np.maximum(R, 1e-9)
        r = np.hypot(x_, y_)
        cos, sin = x_/r, y_/r

        # shear
        gamma_1_ = (1-t)*(alpha_x_*cos - alpha_y_*sin)/r - kappa_*(cos*cos*2 - 1)
        gamma_2_ = (1-t)*(alpha_y_*cos + alpha_x_*sin)/r - kappa_*(sin*cos*2)
        gamma_1_ = np.nan_to_num(gamma_1_, neginf=-1e10, posinf=1e10)
        gamma_2_ = np.nan_to_num(gamma_2_, neginf=-1e10, posinf=1e10)

        # hessian derivatives, still oriented along major axis?
        #H_xx = kappa + gamma_1_
        #H_yy = kappa - gamma_1_
        #H_xy = gamma_2_
        #H_yx = H_xy

        kappa = kappa_
        gamma_1 = np.cos(2 * phi_) * gamma_1_ - np.sin(2 * phi_) * gamma_2_
        gamma_2 = np.sin(2 * phi_) * gamma_1_ + np.cos(2 * phi_) * gamma_2_

        H_xx = kappa + gamma_1
        H_yy = kappa - gamma_1
        H_xy = gamma_2
        H_yx = H_xy

        return H_xx, H_xy, H_yx, H_yy


class ExternalShear(BaseMassProfile):
    """
    Coordinates of the origin for the external shear profile are assumed to be (0., 0.).
    """ 

    _template_class = TemplateExternalShear()

    def potential(self, x, y, gamma_ext=0., phi_ext=0.):
        phi_ext_ = util.eastofnorth2normalradians(phi_ext)
        r, phi = util.cartesian2polar(x, y)
        return 1. / 2 * gamma_ext * r**2 * np.cos(2. * (phi - phi_ext_))

    def deflection(self, x, y, gamma_ext=0., phi_ext=0.):
        phi_ext_ = util.eastofnorth2normalradians(phi_ext)
        gamma1 = gamma_ext * np.cos(2.*phi_ext_)
        gamma2 = gamma_ext * np.sin(2.*phi_ext_)
        x_ = x # no shift
        y_ = y # no shift
        a_x = gamma1 * x_ + gamma2 * y_
        a_y = gamma2 * x_ - gamma1 * y_
        return a_x, a_y

    def convergence(self, x, y, gamma_ext=0., phi_ext=0.):
        return np.zeros_like(x)

    def hessian(self, x, y, gamma_ext=0., phi_ext=0.):
        kappa = 0.
        phi_ext_ = util.eastofnorth2normalradians(phi_ext)
        gamma1 = gamma_ext * np.cos(2.*phi_ext_)
        gamma2 = gamma_ext * np.sin(2.*phi_ext_)
        H_xx = kappa + gamma1
        H_yy = kappa - gamma1
        H_xy = gamma2
        H_yx = H_xy
        return H_xx, H_xy, H_yx, H_yy
    

class ConvergenceSheet(BaseMassProfile):
    """
    Coordinates of the origin for the convergence sheet are assumed to be (0., 0.).
    """

    _template_class = TemplateConvergenceSheet()

    def potential(self, x, y, kappa_s=0.):
        x_ = x # no shift
        y_ = y # no shift
        r_ = np.hypot(x_, y_)
        return 0.5 * kappa_s * r_**2
    
    def deflection(self, x, y, kappa_s=0.):
        x_ = x # no shift
        y_ = y # no shift
        return x_ * kappa_s, y_ * kappa_s

    def convergence(self, x, y, kappa_s=0.):
        return np.full_like(x, kappa_s)

    def hessian(self, x, y, kappa_s=0.):
        kappa = np.full_like(x, kappa_s)
        gamma1 = 0.
        gamma2 = 0.
        H_xx = kappa + gamma1
        H_yy = kappa - gamma1
        H_xy = gamma2
        H_yx = H_xy
        return H_xx, H_xy, H_yx, H_yy



__author__ = 'aymgal'


from coolest.template.classes.base import APIBaseObject

__all__ = [
    'PosteriorStatistics',
    'Prior', 
    'GaussianPrior', 
    'LogNormalPrior',
    'UniformPrior',
]

PRIOR_SUPPORTED_CHOICES = [
    'GaussianPrior', 
    'LogNormalPrior',
    'UniformPrior',
]


class PosteriorStatistics(APIBaseObject):
    """Object that holds low order statistics of the posterior distribution 
    of the parameter.

    Parameters
    ----------
    mean : float, optional
        Mean of the distribution, by default None
    median : float, optional
        Median of the distribution, by default None
    percentile_16th : float, optional
        16th percentile of the distribution, by default None
    percentile_84th : float, optional
        84th percentile of the distribution, by default None
    """
    
    def __init__(self, mean=None, median=None, 
                 percentile_16th=None, percentile_84th=None):
        self.mean = mean
        self.median = median
        self.percentile_16th = percentile_16th
        self.percentile_84th = percentile_84th


class Prior(APIBaseObject):
    """Generic class for a prior distribution that can be assigned
    to a parameter.

    Parameters
    ----------
    ptype : str, optional
        Type of prior, typically class name of one of the Prior class defined
        in this module, by default None
    """
    
    def __init__(self, ptype=None, **kwargs):
        self.type = ptype
        for key, value in kwargs.items():
            setattr(self, key, value)
        super().__init__()


class GaussianPrior(Prior):
    """Gaussian prior.

    Parameters
    ----------
    mean : float, optional
        Mean value, by default None
    width : float, optional
        Width (standard deviation), by default None
    """
    
    def __init__(self, mean=None, width=None):
        super().__init__(self.__class__.__name__, 
                         mean=mean, width=width)


class LogNormalPrior(Prior):
    """Log-Normal prior.

    Parameters
    ----------
    mean : float, optional
        Mean value, by default None
    width : float, optional
        Width, by default None
    """
    
    def __init__(self, mean=None, width=None):
        super().__init__(self.__class__.__name__, 
                         mean=mean, width=width)


class UniformPrior(Prior):
    """Uniform prior.

    Parameters
    ----------
    min_value : float, optional
        Lower bound, by default None
    max_value : float, optional
        Upper bound, by default None
    """
    
    def __init__(self, min_value=None, max_value=None):
        super().__init__(self.__class__.__name__, 
                         min_value=min_value, max_value=max_value)



__author__ = 'aymgal'

from coolest.template.classes.lensing_entity import LensingEntity
from coolest.template.classes.mass_light_model import MassModel
from coolest.template.classes import util


class MassField(LensingEntity):
    """Generic field of massive objects, for instance an external shear field.

    Parameters
    ----------
    name : str
        Name associated to that shear component.
    redshift : float
        Redshift associated to that shear component, if needed.
    mass_model : MassModel, optional
        Mass model of the field, by default None
    """

    def __init__(self,
                 name: str,
                 redshift: float,
                 mass_model: MassModel = None) -> None:
        super().__init__(name, redshift, mass_model=mass_model)
        


__author__ = 'aymgal'

from coolest.template.classes.psf import PSF
from coolest.template.classes.base import APIBaseObject


class Instrument(APIBaseObject):
    """Defines the instrument used for the observation.
    This includes the name of the telescope and detector, the filter, 
    the pixel size, the readout noise and the point spread function (PSF).

    Parameters
    ----------
    pixel_size : float
        Size in arcseconds of a single detector pixel.
    name : str, optional
        Name of the instrument, by default ""
    band : str, optional
        Name of the filter, by default ""
    readout_noise : float, optional
        Readout noise (in electrons), by default 0.
    psf : PSF, optional
        Instance of PSF object, by default None
    """

    def __init__(self,
                 pixel_size: float,
                 name: str = "",
                 band: str = "",
                 readout_noise: float = 0.,
                 psf: PSF = None) -> None:
        self.name = name
        self.band = band
        self.pixel_size = pixel_size
        self.readout_noise = readout_noise
        if psf is None:
            psf = PSF()
        self.psf = psf
        super().__init__()



import setuptools
import os

name = 'coolest'

release_info = {}
infopath = os.path.abspath(os.path.join(os.path.dirname(__file__),
                           name, '__init__.py'))
with open(infopath) as open_file:
    exec(open_file.read(), release_info)

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

# Python version
python_requires = '>=3.7'

# Required packages
install_requires = [
    'numpy>=1.20.0',
    'scipy>=1.6.3',
    'jsonpickle>=2.0.0',
    'astropy>=4.2.0',
]

# Optional packages
install_optional = [
    'matplotlib>=3.7.0',    # for plotting
    'lenstronomy>=1.11.0',  # for shapelets light profiles
    'ipython',              # for running example notebooks
    'ipykernel',            # notebooks in custom environment
    'getdist>=1.3.2',       # for making corner plots
]

version = release_info['__version__']

setuptools.setup(
    name=name,
    author=release_info['__author__'],
    author_email=release_info['__email__'],
    version=version,
    url=release_info['__url__'],
    download_url=f"https://github.com/aymgal/coolest/archive/refs/tags/v{version}.tar.gz",
    packages=setuptools.find_packages(),
    license=release_info['__license__'],
    description=release_info['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords=["coolest", "lensing", "gravitation", "astrophysics"],

    python_requires=python_requires,
    install_requires=install_requires,

    extras_require={
        "opt": install_optional,  # installable via `pip install coolest[opt]`
    },

    setup_requires=['pytest-runner',],
    tests_require=['pytest', 'pytest-cov', 'pytest-pep8'],
)



# Single parameter of a profile

from typing import List

from coolest.template.classes.base import APIBaseObject
from coolest.template.classes.probabilities import Prior, PosteriorStatistics
from coolest.template.classes.grid import PixelatedRegularGrid, IrregularGrid

import numpy as np


__all__ = [
    'Parameter',
    'NonLinearParameter', 
    'LinearParameter', 
    'HyperParameter',
    'LinearParameterSet',
    'NonLinearParameterSet',
    'PixelatedRegularGridParameter',
    'IrregularGridParameter',
]


class DefinitionRange(APIBaseObject):
    """Defines the interval over which a parameter is defined.

    Parameters
    ----------
    min_value : (int, float), optional
        Lower bound of the interval (inclusive), by default None
    max_value : (int, float), optional
        Upper bound of the interval (inclusive), by default None
    """
    
    def __init__(self, min_value=None, max_value=None):
        self.min_value = min_value
        self.max_value = max_value


class PointEstimate(APIBaseObject):
    """Define a point in the parameter space.

    Parameters
    ----------
    value : float, optional
        Value of the parameter, by default None
    """
    
    def __init__(self, value=None):
        self.value = value


class Parameter(APIBaseObject):
    """Base class of a generic model parameter.

    Parameters
    ----------
    documentation : str
        Short description of the parameter.
    definition_range : DefinitionRange, optional
        Interval over which the parameter is defined, by default None
    units : str, optional
        Unit of the parameter, if any, by default None
    fixed : bool, optional
        If True, the parameter is considered fixed 
        (i.e. should not be, or has not be optimized), by default False
    point_estimate : PointEstimate, optional
        Point-estimate value of the parameter, by default None
    posterior_stats : PosteriorStatistics, optional
        Summary statistics of the marginalized posterior 
        distribution of the parameter, by default None
    prior : Prior, optional
        Prior assigned the parameter, if any, by default None
    latex_str : str, optional
        LaTeX representation of the parameter, by default None
    
    #TODO: for parameters like orientation / axis ratio, add a class method to compute
                 related quantities like ellipticity parameters.
    """
    def __init__(self, 
                 documentation: str, 
                 definition_range: DefinitionRange = None,
                 units: str = None,
                 fixed: bool = False,
                 point_estimate: PointEstimate = None,
                 posterior_stats: PosteriorStatistics = None,
                 prior: Prior = None,
                 latex_str: str = None) -> None:
        self.documentation = documentation
        self.units = units
        self.definition_range = definition_range
        self.fixed = fixed
        if not isinstance(point_estimate, PointEstimate):
            self.point_estimate = PointEstimate(point_estimate)
        else:
            self.point_estimate = point_estimate
        if posterior_stats is None:
            posterior_stats = PosteriorStatistics()
        self.posterior_stats = posterior_stats
        if prior is None:
            prior = Prior()
        self.prior = prior
        self.latex_str = latex_str
        self.id = None
        super().__init__()
        
    def set_point_estimate(self, point_estimate):
        """Set the point estimate value of the parameter.

        Parameters
        ----------
        point_estimate : int, float, list, PointEstimate
            Parameter value, or directly a PointEstimate instance.

        Raises
        ------
        ValueError
            If the provided point_estimate has not a supported type.
        ValueError
            If the parameter value is below its minimum allowed value.
        ValueError
            If the parameter value is above its maximum allowed value.
        """
        if isinstance(point_estimate, (float, int, list)):
            self.point_estimate = PointEstimate(value=point_estimate)
        elif isinstance(point_estimate, tuple):
            self.point_estimate = PointEstimate(value=list(point_estimate))
        elif isinstance(point_estimate, np.ndarray):
            self.point_estimate = PointEstimate(value=point_estimate.tolist())
        elif isinstance(point_estimate, PointEstimate):
            self.point_estimate = point_estimate
        else:
            raise ValueError("Parameter prior must be either a PointEstimate instance "
                             "or a single number (float or int) or an array (tuple, list or ndarray).")
        if self.point_estimate.value is not None:
            val = self.point_estimate.value
            min_val = self.definition_range.min_value
            max_val = self.definition_range.max_value
            if min_val is not None and np.any(np.asarray(val) < np.asarray(min_val)):
                raise ValueError(f"Value cannot be smaller than {self.definition_range.min_value}.")
            if max_val is not None and np.any(np.asarray(val) > np.asarray(max_val)):
                raise ValueError(f"Value cannot be larger than {self.definition_range.max_value}.")

    def remove_point_estimate(self):
        """Remove the current point estimate of the parameter.
        """
        self.point_estimate = PointEstimate()

    def set_posterior(self, posterior_stats):
        """Set the posterior statistics of the parameter.

        Parameters
        ----------
        posterior_stats : PosteriorStatistics
            Instance of the PosteriorStatistics object.

        Raises
        ------
        ValueError
            If the argument is not a PosteriorStatistics instance.
        """
        if not isinstance(posterior_stats, PosteriorStatistics):
            raise ValueError("Parameter prior must be a PosteriorStatistics instance.")
        self.posterior_stats = posterior_stats

    def remove_posterior(self):
        """Remove the current posterior statistics of the parameter.
        """
        self.posterior_stats = PosteriorStatistics()

    def set_prior(self, prior):
        """Associate a prior distribution to the parameter.

        Parameters
        ----------
        prior : Prior
            Instance of Prior object.

        Raises
        ------
        ValueError
            If the argument is not a Prior instance.
        """
        if not isinstance(prior, Prior):
            raise ValueError("Parameter prior must be a Prior instance.")
        self.prior = prior

    def remove_prior(self):
        """Remove the current posterior statistics of the parameter.
        """
        self.prior = Prior()

    def fix(self):
        """Set the fixed attribute to True, marking it as fixed."""
        if self.point_estimate.value is None:
            raise ValueError("Cannot fix parameter as no point estimate value has been set.")
        self.fixed = True

    def unfix(self):
        """Set the fixed attribute to False, marking as free to vary"""
        self.fixed = False


class NonLinearParameter(Parameter):
    """Define a non-linear parameter of a lens model
    
    Warning: this class may be removed in the future.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LinearParameter(Parameter):
    """Define a hyper-parameter of a lens model
    
    Warning: this class may be removed in the future, as it has adds unncessary abstraction level.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class HyperParameter(Parameter):
    """Define a hyper-parameter of a model"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)        


class ParameterSet(Parameter):
    """Typically for analytical basis sets"""

    def __init__(self, *args, **kwargs) -> None:
        if 'point_estimate' not in kwargs or kwargs['point_estimate'] is None:
            kwargs['point_estimate'] = []
        if not isinstance(kwargs['point_estimate'], list):
            raise ValueError("For any ParameterSet, `point_estimate` must be a list of values.")
        super().__init__(*args, **kwargs)
        self.num_values = len(self.point_estimate.value)


class LinearParameterSet(ParameterSet):
    """Typically for analytical basis sets.

    Warning: this class may be removed in the future, as it has adds unncessary abstraction level.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        
        
class NonLinearParameterSet(ParameterSet):
    """Typically for position of point sources.
    
    Warning: this class may be removed in the future, as it has adds unncessary abstraction level."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class PixelatedRegularGridParameter(PixelatedRegularGrid):
    """Typically for pixelated profiles"""
    # TODO: implement .fixed attribute following the analytical Parameter interface

    def __init__(self, documentation, **kwargs_grid) -> None:
        self.documentation = documentation
        super().__init__(**kwargs_grid)


class IrregularGridParameter(IrregularGrid):
    """Typically for pixelated profiles"""
    # TODO: implement .fixed attribute following the analytical Parameter interface

    def __init__(self, documentation, **kwargs_grid) -> None:
        self.documentation = documentation
        super().__init__(**kwargs_grid)
        


# Some variables useful to access

from coolest.template import standard
from coolest.template.classes import lensing_entity
from coolest.template.classes.profiles import light
from coolest.template.classes.profiles import mass
from coolest.template.classes import noise
from coolest.template.classes import psf
from coolest.template.classes import likelihood_list
from coolest.template.classes import probabilities
# from coolest.template.classes.regularizations import pixelated as pixel_regul


all_supported_choices = {
    'modes': standard.SUPPORTED_MODES,
    'lensing_entities': lensing_entity.SUPPORTED_CHOICES,
    'light_profiles': light.SUPPORTED_CHOICES,
    'mass_profiles': mass.SUPPORTED_CHOICES,
    # 'pixel_regularizations': pixel_regul.SUPPORTED_CHOICES,
    'likelihoods': likelihood_list.SUPPORTED_CHOICES,
    'noise_types': noise.SUPPORTED_CHOICES,
    'psf_types': psf.SUPPORTED_CHOICES,
    'prior_types': probabilities.PRIOR_SUPPORTED_CHOICES
}



__author__ = 'aymgal', 'Giorgos Vernardos'


import numpy as np
import warnings
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import ticker
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.spatial import Voronoi, voronoi_plot_2d
from matplotlib.colors import Normalize, LogNorm, TwoSlopeNorm
from matplotlib.cm import ScalarMappable


def plot_voronoi(ax, x, y, z, neg_values_as_bad=False, 
                 norm=None, cmap=None, zmin=None, zmax=None, 
                 edgecolor=None, zorder=1):

    if cmap is None:
        cmap = plt.get_cmap('inferno')
    if norm is None:
        if zmin is None:
            zmin = np.min(z)
        if zmax is None:
            zmax = np.max(z)
        norm = Normalize(zmin, zmax)

    # get voronoi regions
    voronoi_points = np.column_stack((x,y))
    vor = Voronoi(voronoi_points)
    new_regions, vertices = voronoi_finite_polygons_2d(vor)
    
    # get cell colors
    m = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap)

    # plot voronoi points
    #point_colors = [ m.to_rgba(v) for v in z ]
    #ax.scatter(voronoi_points[:,0],voronoi_points[:,1],c=point_colors)

    # plot voronoi cells
    for i, region in enumerate(new_regions):
        polygon = vertices[region]
        z_i = z[i]
        if neg_values_as_bad is True and z_i < 0.:
            cell_color = m.to_rgba(np.nan)
        else:
            cell_color = m.to_rgba(z_i)
        ax.fill(*zip(*polygon), facecolor=cell_color, edgecolor=edgecolor, zorder=zorder)
    return m


def voronoi_finite_polygons_2d(vor,radius=None):
    """
    Reconstruct infinite voronoi regions in a 2D diagram to finite
    regions.
    This function is taken from: https://gist.github.com/pv/8036995

    Parameters
    ----------
    vor : Voronoi
        Input diagram
    radius : float, optional
        Distance to 'points at infinity'.

    Returns
    -------
    regions : list of tuples
        Indices of vertices in each revised Voronoi regions.
    vertices : list of tuples
        Coordinates for revised Voronoi vertices. Same as coordinates
        of input vertices, with 'points at infinity' appended to the
        end.
    """

    if vor.points.shape[1] != 2:
        raise ValueError("Requires 2D input")

    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max()

    # Construct a map containing all ridges for a given point
    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    # Reconstruct infinite regions
    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            # finite region
            new_regions.append(vertices)
            continue

        # reconstruct a non-finite region
        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                # finite ridge: already in the region
                continue

            # Compute the missing endpoint of an infinite ridge

            t = vor.points[p2] - vor.points[p1] # tangent
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])  # normal

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        # sort region counterclockwise
        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:,1] - c[1], vs[:,0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        # finish
        new_regions.append(new_region.tolist())
        
    return new_regions, np.asarray(new_vertices)


def std_colorbar(mappable, label=None, fontsize=12, label_kwargs={}, **colorbar_kwargs):
    cb = plt.colorbar(mappable, **colorbar_kwargs)
    if label is not None:
        colorbar_kwargs.pop('label', None)
        cb.set_label(label, fontsize=fontsize, **label_kwargs)
    return cb

def std_colorbar_residuals(mappable, res_map, vmin, vmax, label=None, fontsize=12, 
                           label_kwargs={}, **colorbar_kwargs):
    if res_map.min() < vmin and res_map.max() > vmax:
        cb_extend = 'both'
    elif res_map.min() < vmin:
        cb_extend = 'min'
    elif res_map.max() > vmax:
        cb_extend = 'max'
    else:
        cb_extend = 'neither'
    colorbar_kwargs.update({'extend': cb_extend})
    return std_colorbar(mappable, label=label, fontsize=fontsize, 
                        label_kwargs=label_kwargs, **colorbar_kwargs)

def nice_colorbar(mappable, ax=None, position='right', pad=0.1, size='5%', label=None, fontsize=12, 
                  invisible=False, max_nbins=None,
                  divider_kwargs={}, colorbar_kwargs={}, label_kwargs={}):
    divider_kwargs.update({'position': position, 'pad': pad, 'size': size})
    if ax is None:
        ax = mappable.axes
    divider = make_axes_locatable(ax)
    cax = divider.append_axes(**divider_kwargs)
    if invisible:
        cax.axis('off')
        return None
    cb = plt.colorbar(mappable, cax=cax, **colorbar_kwargs)
    if label is not None:
        colorbar_kwargs.pop('label', None)
        cb.set_label(label, fontsize=fontsize, **label_kwargs)
    if position == 'top':
        cax.xaxis.set_ticks_position('top')
    if max_nbins is not None:
        cb.locator = ticker.MaxNLocator(nbins=max_nbins)
        cb.update_ticks()
    return cb

def nice_colorbar_residuals(mappable, res_map, vmin, vmax, ax=None, position='right', pad=0.1, size='5%', 
                            invisible=False, label=None, fontsize=12,
                            divider_kwargs={}, colorbar_kwargs={}, label_kwargs={}):
    if res_map.min() < vmin and res_map.max() > vmax:
        cb_extend = 'both'
    elif res_map.min() < vmin:
        cb_extend = 'min'
    elif res_map.max() > vmax:
        cb_extend = 'max'
    else:
        cb_extend = 'neither'
    colorbar_kwargs.update({'extend': cb_extend})
    return nice_colorbar(mappable, ax=ax, position=position, pad=pad, size=size, label=label, fontsize=fontsize,
                  invisible=invisible, colorbar_kwargs=colorbar_kwargs, label_kwargs=label_kwargs,
                  divider_kwargs=divider_kwargs)

def cax_colorbar(fig, cax, norm=None, cmap=None, mappable=None, label=None, fontsize=12, orientation='horizontal', label_kwargs={}):
    if mappable is None:
        mappable = ScalarMappable(norm=norm, cmap=cmap)
    cb = fig.colorbar(mappable=mappable, orientation=orientation, cax=cax)
    if label is not None:
        cb.set_label(label, fontsize=fontsize, **label_kwargs)

def plot_regular_grid(ax, image_, neg_values_as_bad=False, xylim=None, **imshow_kwargs):
    if neg_values_as_bad:
        image = np.copy(image_)
        image[image < 0] = np.nan
    else:
        image = image_
    im = ax.imshow(image, **imshow_kwargs)
    im.set_rasterized(True)
    set_xy_limits(ax, xylim)
    ax.xaxis.set_major_locator(plt.MaxNLocator(3))
    ax.yaxis.set_major_locator(plt.MaxNLocator(3))
    return ax, im

def plot_irregular_grid(ax, points, xylim, neg_values_as_bad=False,
                            norm=None, cmap=None, plot_points=False):
    x, y, z = points
    im = plot_voronoi(ax, x, y, z, neg_values_as_bad=neg_values_as_bad, 
                      norm=norm, cmap=cmap, zorder=1)
    ax.set_aspect('equal', 'box')
    set_xy_limits(ax, xylim)
    ax.xaxis.set_major_locator(plt.MaxNLocator(3))
    ax.yaxis.set_major_locator(plt.MaxNLocator(3))
    if plot_points:
        ax.scatter(x, y, s=5, c='white', marker='.', alpha=0.4, zorder=2)
    return ax, im

def set_xy_limits(ax, xylim):
    if xylim is None: return  # do nothing
    if isinstance(xylim, (int, float)):
        xylim_ = [-xylim, xylim, -xylim, xylim]
    elif isinstance(xylim, (list, tuple)) and len(xylim) == 2:
        xylim_ = [xylim[0], xylim[1], xylim[0], xylim[1]]
    elif not isinstance(xylim, (list, tuple)) and len(xylim) != 4:
        raise ValueError("`xylim` argument should be a single number, a 2-tuple or a 4-tuple.")
    else:
        xylim_ = xylim
    ax.set_xlim(xylim_[0], xylim_[1])
    ax.set_ylim(xylim_[2], xylim_[3])

def panel_label(ax, text, color, fontsize, alpha=0.8, loc='upper left'):
    if loc == 'upper left':
        x, y, ha, va = 0.03, 0.97, 'left', 'top'
    elif loc == 'lower left':
        x, y, ha, va = 0.03, 0.03, 'left', 'bottom'
    elif loc == 'upper right':
        x, y, ha, va = 0.97, 0.97, 'right', 'top'
    elif loc == 'lower right':
        x, y, ha, va = 0.97, 0.03, 'right', 'bottom'
    ax.text(x, y, text, color=color, fontsize=fontsize, alpha=alpha, 
            ha=ha, va=va, transform=ax.transAxes)



__author__ = 'aymgal', 'lynevdv', 'Giorgos Vernardos'


import numpy as np
from scipy import interpolate

from coolest.template.classes.profiles.light import (Sersic as TemplateSersic,
                                                     Shapelets as TemplateShapelets,
                                                     PixelatedRegularGrid as TemplatePixelatedRegularGrid,
                                                     IrregularGrid as TemplateIrregularGrid)
from coolest.api.profiles import util


class BaseLightProfile(object):

    _template_class = None
    _units = None

    def surface_brightness(self, **params):
        raise NotImplementedError(f"The method surface_brightness() is not defined "
                                  f"for profile '{self.__class__.__name__}'")

    def evaluate_surface_brightness(self, **params):
        raise NotImplementedError(f"The method evaluate_surface_brightness() is not defined "
                                  f"for profile '{self.__class__.__name__}'")

    def get_extent(self):
        return None

    def get_coordinates(self):
        return None

    @property
    def units(self):
        if not hasattr(self, '_units'):
            raise ValueError(f"Units for profile {self.__class__.__name__} must be specified")
        if self._units not in ('per_pix', 'per_ang'):
            raise ValueError(f"Unsupported units type {self._units}")
        return self._units

    @property
    def template_class(self):
        if self._template_class is None:
            raise RuntimeError("No template class has been set by light profile class")
        return self._template_class

    @property
    def type(self):
        return self.template_class.type

    @property
    def parameter_names(self):
        return list(self.template_class.parameters.keys())


class Sersic(BaseLightProfile):

    """Elliptical Sersic"""

    _units = 'per_ang'
    _template_class = TemplateSersic()

    def surface_brightness(self, I_eff=1., theta_eff=2., n=4., phi=0., q=1., center_x=0., center_y=0.):
        raise ValueError("Sersic surface brightness can only be evaluated")

    def evaluate_surface_brightness(self, x, y, I_eff=1., theta_eff=2., n=4., phi=0., q=1., center_x=0., center_y=0.):
        """Returns the surface brightness at the given position (x, y)"""
        phi_ = util.eastofnorth2normalradians(phi)
        x_t, y_t = util.shift_rotate_elliptical(x, y, phi_, q, center_x, center_y)
        bn = 1.9992*n - 0.3271
        return I_eff * np.exp( - bn * ( (np.sqrt(x_t**2+y_t**2) / theta_eff )**(1./n) -1. ) )


class Shapelets(BaseLightProfile):

    """Elliptical Sersic
    
    The implementation uses functions from `lenstronomy` (:cite:t:`lenstronomy2018`, :cite:t:`lenstronomy2021`), based on the developments of :cite:t:`Refregier2003`.
    """

    _units = 'per_ang'
    _template_class = TemplateShapelets()

    def __init__(self):
        from lenstronomy.LightModel.Profiles.shapelets import ShapeletSet
        self._backend = ShapeletSet()

    def surface_brightness(self, amps=0, n_max=0, beta=0, center_x=0, center_y=0):
        raise ValueError("Surface brightness of a set of shapelets can only be evaluated")

    def evaluate_surface_brightness(self, x, y, amps=0, n_max=0, beta=0, center_x=0, center_y=0):
        """Returns the surface brightness at the given position (x, y)"""
        x_, y_ = x.flatten(), y.flatten()
        flux = self._backend.function(x_, y_, amps, n_max, beta, center_x=center_x, center_y=center_y)
        return flux.reshape(*x.shape)


class PixelatedRegularGrid(BaseLightProfile):

    """Pixelated profile on a regular grid"""

    _units = 'per_pix'
    _template_class = TemplatePixelatedRegularGrid()

    def __init__(self, field_of_view_x, field_of_view_y, num_pix_x, num_pix_y,
                 interpolation_method='cubic'):
        if num_pix_x == 0 or num_pix_y == 0:
            raise ValueError("Light profile defined on regular grid has zero pixels")
        self._fov_x = field_of_view_x
        self._fov_y = field_of_view_y
        self._nx = num_pix_x
        self._ny = num_pix_y
        self._shape = (num_pix_x, num_pix_y)
        self._pix_scl_x = np.abs(self._fov_x[0] - self._fov_x[1]) / self._nx
        self._pix_scl_y = np.abs(self._fov_y[0] - self._fov_y[1]) / self._ny
        self._interp_method = interpolation_method

    def surface_brightness(self, pixels=None):
        """Returns the surface brightness pixels"""
        if pixels is None:
            return np.zeros(self._shape)
        return pixels

    def evaluate_surface_brightness(self, x, y, pixels=None):
        coordinates = self.get_coordinates()
        points = coordinates.pixel_axes
        interp = util.CartesianGridInterpolator(points, pixels, method=self._interp_method)
        points_eval = np.array([y.ravel(), x.ravel()]).T
        pixels_eval = interp(points_eval).reshape(*x.shape)
        return pixels_eval

    def get_extent(self):
        coordinates = self.get_coordinates()
        return coordinates.plt_extent

    def get_coordinates(self):
        from coolest.api.util import get_coordinates_from_regular_grid
        return get_coordinates_from_regular_grid(self._fov_x, self._fov_y, self._nx, self._ny)


class IrregularGrid(BaseLightProfile):

    """Pixelated profile on an irregular grid of points {x, y, z}"""

    _units = 'per_pix'
    _template_class = TemplateIrregularGrid()

    def __init__(self, field_of_view_x, field_of_view_y, num_pix,
                 interpolation_method='cubic'):
        self._fov_x = field_of_view_x
        self._fov_y = field_of_view_y
        self._n = num_pix
        self._interp_method = interpolation_method

    def surface_brightness(self, x=None, y=None, z=None):
        """Returns the surface brightness pixels"""
        if x is None or y is None or z is None:
            return np.zeros(self._n), np.zeros(self._n), np.zeros(self._n)
        return x, y, z

    def evaluate_surface_brightness(self, x_eval, y_eval, x=None, y=None, z=None):
        z_eval = interpolate.griddata((x, y), z, (x_eval, y_eval), 
                                      method=self._interp_method)
        return z_eval

    def get_extent(self):
        return [
            self._fov_x[0], 
            self._fov_x[1], 
            self._fov_y[0], 
            self._fov_y[1]
        ]



__author__ = 'aymgal'


from astropy.coordinates import SkyCoord

from coolest.template.classes.base import APIBaseObject


class CoordinatesOrigin(APIBaseObject):
    """Defines the origin of the coordinate system, given in RA-Dec.
    That origin is then assumed to lie at the center of a data pixel.

    Parameters
    ----------
    ra : str, optional
        Right ascension (hms format), by default '0h0m0.0s'
    dec : str, optional
        Declination (dms format), by default '0d0m0.0s'
    """

    def __init__(self, 
                 ra: str = '0h0m0.0s',
                 dec: str = '0d0m0.0s') -> None:
        self.ra, self.dec = self._check_sky_coord(ra, dec)
        super().__init__()

    def _check_sky_coord(self, origin_ra, origin_dec):
        # creates astropy object to check everything is good
        sky_coord = SkyCoord(origin_ra, origin_dec, frame='icrs')
        return sky_coord.to_string(style='hmsdms').split(' ')


# class Coordinates(APIBaseObject):

#     _orientations_ra = ['left', 'right']
#     _orientations_dec = ['top', 'bottom']
#     _origins = ['center', 'bottom-left']
#     _angles = ['east-to-north', 'west-to-north', 'east-to-south', 'west-to-south']
#     _coord_in_pixel = ['center', 'bottom-left']

#     def __init__(self, 
#                  orientation_ra: str = 'left', 
#                  orientation_dec: str = 'top',
#                  origin_position: str = 'center',
#                  angles_orientation: str = 'east-to-north',
#                  coordinate_in_pixel: str = 'center',
#                  ) -> None:
#         if orientation_ra not in self._orientations_ra:
#             raise ValueError(f"RA orientation can only be in {self._orientations_ra}.")
#         self.orientation_ra = orientation_ra
#         if orientation_dec not in self._orientations_dec:
#             raise ValueError(f"Dec orientation can only be in {self._orientations_dec}.")
#         self.orientation_dec = orientation_dec
#         if origin_position not in self._origins:
#             raise ValueError(f"Position of the origin can only be in {self._origins}.")
#         self.origin_position = origin_position
#         if angles_orientation not in self._angles:
#             raise ValueError(f"Angle orientation can only be in {self._angles}.")
#         self.angles_orientation = angles_orientation
#         if coordinate_in_pixel not in self._coord_in_pixel:
#             raise ValueError(f"Coordinate position inside a given pixel "
#                              "can only be in {self._coord_in_pixel}.")
#         self.coordinate_in_pixel = coordinate_in_pixel
#         super().__init__()

#     def update_with_instrument(self, instrument):
#         pixel_size = instrument.pixel_size
#         sign_ra  = (self.orientation_ra == 'left')
#         sign_dec = (self.orientation_dec == 'top')
#         self.pix2angle_matrix = [
#             [sign_ra * pixel_size, 0],
#             [0, sign_dec * pixel_size],
#         ]
#         if self.origin_position == 'center':
#             half_fov_ra = instrument.field_of_view_ra / 2.
#             half_fov_dec = instrument.field_of_view_ra / 2.
#             self.ra_at_xy_0 = - half_fov_ra + pixel_size / 2.
#             self.dec_at_xy_0 = - half_fov_dec + pixel_size / 2.
#         else:
#             self.ra_at_xy_0 = 0.
#             self.dec_at_xy_0 = 0.



__author__ = 'aymgal'

from types import ModuleType
from typing import Tuple

from coolest.template.classes.base import APIBaseObject


class ProfileList(list, APIBaseObject):

    def __init__(self, 
                 profiles_module: ModuleType,
                 *profile_names: Tuple[str]) -> None:
        """Generic list of model (mass or light) profile.
        Note that this class is not meant to be instantiated by the user.

        Parameters
        ----------
        profiles_module : mass or light profile submodule
            Submodule in coolest.template.classes.profiles defining 
            the python classes corresponding to each model profile.
        *profile_names: list of profile names, which must be class names
        defined in the profiles_module submodule.

        Raises
        ------
        ValueError
            If profile is not supported.
        """
        profiles = []
        for name in profile_names:
            if name not in profiles_module.SUPPORTED_CHOICES:
                raise ValueError(f"Profile '{name}' is not supported.")
            profile_class = getattr(profiles_module, name)
            profiles.append(profile_class())
        list.__init__(self, profiles)

    def num_pixelated_profiles(self):
        pass  #TODO



# imports all classes that are need to be filled

from coolest.template.classes.grid import PixelatedRegularGrid,IrregularGrid
from coolest.template.classes.fits_file import FitsFile
from coolest.template.classes.observation import Observation
from coolest.template.classes.psf import *
from coolest.template.classes.noise import *
from coolest.template.classes.instrument import Instrument
from coolest.template.classes.observation import Observation
from coolest.template.classes.coordinates import CoordinatesOrigin
from coolest.template.classes.cosmology import Cosmology
from coolest.template.classes.galaxy import Galaxy
from coolest.template.classes.mass_field import MassField
from coolest.template.classes.lensing_entity_list import LensingEntityList
from coolest.template.classes.mass_light_model import MassModel, LightModel
from coolest.template.classes.probabilities import Prior, PosteriorStatistics

